/* src/ocsp_response.c
 * ============================================================
 * OCSP Responder
 * (c) 2001-2006 by Massimiliano Pala and OpenCA Group
 *     OpenCA License software
 * ============================================================
 */
 
#include "general.h"

pthread_mutex_t sign_mutex = PTHREAD_MUTEX_INITIALIZER;

PKI_X509_OCSP_RESP *make_ocsp_response( PKI_X509_OCSP_REQ *req, 
						OCSPD_CONFIG *conf ) {

	OCSP_CERTID *cid = NULL;

	PKI_X509_OCSP_RESP *resp = NULL;
	PKI_X509_OCSP_REQ_VALUE *req_val = NULL;

	PKI_TOKEN *tk = NULL;
	PKI_X509_CERT *cert = NULL;

	int i, id_count;

	PKI_TIME *thisupd = NULL;
	PKI_TIME *nextupd = NULL;
	if((resp = PKI_X509_OCSP_RESP_new()) == NULL ) {
		PKI_log_err ( "Memory Error" );
		goto end;
	}

	if((req_val = PKI_X509_get_value ( req )) == NULL ) {
		PKI_log_err ( "Memory Error" );
		goto end;
	}

	id_count = OCSP_request_onereq_count( req_val );

	if (id_count <= 0) {
		PKI_log_err("[Thread] ERROR::Request has no onereq "
				"(OpenSSL Error is -> %s)",
				ERR_error_string(ERR_get_error(),NULL));

		PKI_X509_OCSP_RESP_set_status ( resp, 
			PKI_X509_OCSP_RESP_STATUS_MALFORMEDREQUEST );
		goto end;
	}

	/* Let's set the default token for signing */
	tk = conf->token;

	// bs = OCSP_BASICRESP_new();
	if (conf->set_nextUpdate) {
		nextupd = PKI_TIME_new ( (conf->nmin * 60 ) + 
						(conf->ndays * 86400 ));
	}

	/* Examine each certificate id in the request */
	for (i = 0; i < id_count; i++) {
		OCSP_ONEREQ *one = NULL;
		ASN1_INTEGER *serial = NULL;
		CA_LIST_ENTRY *ca = NULL;
		X509_REVOKED *entry = NULL;

		one = OCSP_request_onereq_get0(req_val, i);
		cid = OCSP_onereq_get0_id(one);

		/* Get basic request info */
		OCSP_id_get0_info(NULL, NULL, NULL, &serial, cid);

		if( conf->verbose ) {
			char *s = PKI_INTEGER_get_parsed ( serial );
			PKI_log( PKI_LOG_INFO, 
				"request for certificate serial %s", s);
			if( s ) PKI_Free ( s );
		}

		/* Is this request about our CA? */
		if ((ca = OCSPD_CA_ENTRY_find( conf, cid )) == NULL) {
			if (conf->verbose) {
				char *s = PKI_INTEGER_get_parsed ( serial );
				PKI_log( PKI_LOG_INFO, 
				  "request for non reckognized CA [serial %s]",
					s );
				if( s ) PKI_Free ( s );
			}

			PKI_X509_OCSP_RESP_add ( resp, cid, 
					PKI_OCSP_CERTSTATUS_UNKNOWN,
					NULL, NULL, nextupd,
					0, NULL );

			/* Maybe we could add the serviceLocator extension
 			   we can use the PRQP to find out the server address */

			continue;
		}

		/* If the CA has a specific token, let's use that */
		if( ca->token != NULL ) {
			tk = ca->token;
			if( conf->debug ) {
				PKI_log_debug( "Using the specific token "
					"for the found CA (%s)",ca->token_name);
			}
		} else {
			/* If no specific token but a different server_cert
 			 * is to be used, let's use that instead */
			if ( ca->server_cert ) {
				cert = ca->server_cert;
			}
		}

		if( ca->compromised > 0 ) {
			PKI_X509_OCSP_RESP_add ( resp, cid,
				PKI_OCSP_CERTSTATUS_REVOKED,
				NULL, NULL, nextupd, 
				CRL_REASON_CA_COMPROMISE, NULL );

			continue;
		}

		if( ca->crl_status != CRL_OK ) {
			PKI_X509_OCSP_RESP_set_status ( resp,
				PKI_X509_OCSP_RESP_STATUS_TRYLATER );

			PKI_log_err ("SENT TRYLATER (%s)", 
					get_crl_status_info (ca->crl_status));
			goto end;
		}

		/* Get the entry from the CRL data, if NULL then the
		   certificate is not revoked */
		entry = OCSPD_REVOKED_find( ca, serial );

		/* Sets thisUpdate field to the value of the loaded CRL */
		// thisupd = M_ASN1_TIME_dup(ca->lastUpdate);
		// thisupd = PKI_TIME_dup(ca->lastUpdate);
		thisupd = PKI_TIME_new( 0 );

		if( entry ) {
			OCSP_SINGLERESP *single = NULL;

			int reason = -1;
			void *ext = NULL;

			/* If extensions are found, process them */
			if( entry->extensions ) {
				ASN1_ENUMERATED *asn = NULL;

				if( (asn = X509_REVOKED_get_ext_d2i( entry, 
						NID_crl_reason,NULL,NULL )) != NULL ) {
					reason = ASN1_ENUMERATED_get( asn );
					ASN1_ENUMERATED_free( asn );
				}

				/* Check and add the invalidity date */
				ext = X509_REVOKED_get_ext_d2i( entry, 
					NID_invalidity_date, NULL, NULL );
			}

			if((PKI_X509_OCSP_RESP_add ( resp, cid,
					PKI_OCSP_CERTSTATUS_REVOKED,
					entry->revocationDate, thisupd,
					nextupd, reason, ext )) == PKI_ERR ) {

				PKI_log_err ("Can not generate OCSP response");
			}

			if( reason == CRL_REASON_CERTIFICATE_HOLD ) {
				/* TODO: We might want to add the CrlID
 				   extension to the response */
			}

			if( conf->verbose ) {
				char * s = PKI_INTEGER_get_parsed ( serial );

				PKI_log(PKI_LOG_INFO, 
					"Status for %s is REVOKED", s );
				PKI_Free ( s );
			}
		} else if (ca == NULL ) {
			if( conf->verbose ) {
				char * s = PKI_INTEGER_get_parsed ( serial );
				PKI_log( PKI_LOG_INFO, 
					"status unknown for %ld (unknown CA)",
					s );
				PKI_Free ( s );
			}
			PKI_X509_OCSP_RESP_add ( resp, cid,
				PKI_OCSP_CERTSTATUS_UNKNOWN,
				NULL, thisupd, nextupd, 0, NULL );
		} else {
			if( conf->verbose ) {
				char *s = PKI_INTEGER_get_parsed ( serial );
				PKI_log( PKI_LOG_INFO, "status VALID for %s",s);
				PKI_Free ( s );
			}
			
			PKI_X509_OCSP_RESP_add ( resp, cid,
				PKI_OCSP_CERTSTATUS_GOOD,
				NULL, thisupd, nextupd, 0, NULL );
		}
	}

	if(PKI_X509_OCSP_RESP_copy_nonce( resp, req ) == PKI_ERR ) {
		PKI_log_err ("Can not copy NONCE from request to response");
	}

	if ( !cert && tk ) {
		cert = PKI_TOKEN_get_cert ( tk );
	}

	if ( tk == NULL ) {
		if( conf->debug ) PKI_log_debug ( "CA Token is empty");
		if( conf->token == NULL ) 
				PKI_log_debug ("Default Token is empty!");
	}

	/* It seems that this function is not thread safe!!! */
	if ( tk != NULL ) {

		if ( cert == NULL ) {
			PKI_log_err ("OCSP certificate is empty!");
			if ( resp ) PKI_X509_OCSP_RESP_free ( resp );
			resp = NULL;

			goto end;
		}

		/*
		PKI_log_debug("OCSP Cert Algor: %s",
			PKI_X509_CERT_get_parsed ( tk->cert,
				PKI_X509_DATA_ALGORITHM ));
		*/
		
		PKI_log_debug("Token Algorithm: %s",
			PKI_ALGOR_get_parsed ( tk->algor ));

		/* Sign the response only if we have a valid pkey pointer! */
		if(PKI_X509_OCSP_RESP_sign_tk( resp, tk ) == PKI_ERR ) {
			PKI_log_err("Failed while signing [%s]",
				ERR_error_string(ERR_get_error(), NULL));
			if ( resp ) PKI_X509_OCSP_RESP_free ( resp );
			resp = NULL;

			goto end;
		}

		/* Test Mode: Issues WRONG signatures by flipping the first
 		 * bit in the signature. Use it ONLY for testing OCSP clients
 		 * verify capabilities! */

		if( conf->testmode) {
			PKI_STRING *signature = NULL;

			PKI_log(PKI_LOG_ALWAYS, 
				"WARNING: TestMode (Wrong Signatures): Fipping first bit in Signature");

			// Get The Signature
			signature = PKI_X509_OCSP_RESP_get_data(resp, PKI_X509_DATA_SIGNATURE);
			if(signature)
			{	
				PKI_X509_OCSP_RESP_VALUE *resp_val = NULL;
  			PKI_OCSP_RESP *r = NULL;
				OCSP_BASICRESP *bsrp = NULL;
				
				int i = 0;

				PKI_log_debug("Test Mode: Signature Found!");

				// Flip The First n-Bit of the Signature (n=1)
				for (i=0; i < 1; i++ ) {
					if(ASN1_BIT_STRING_get_bit(signature, i))
    			{
    				ASN1_BIT_STRING_set_bit(signature, i, 0);
    			}
    			else
    			{
        		ASN1_BIT_STRING_set_bit(signature, i, 1);
    			};
				};

				r = resp->value;

				// Now we need to re-encode the basicresp
		  	resp_val = r->resp;
				bsrp = r->bs;

				if( resp_val->responseBytes ) 
					OCSP_RESPBYTES_free(resp_val->responseBytes);

				if (!(resp_val->responseBytes = OCSP_RESPBYTES_new())) {
					PKI_log_err("Memory Error, aborting signature!");
 					return PKI_ERR;
				}

				resp_val->responseBytes->responseType =
					OBJ_nid2obj(NID_id_pkix_OCSP_basic);

				if( bsrp ) {
					/* Now add the encoded data to the request bytes */
					if (!ASN1_item_pack(bsrp, ASN1_ITEM_rptr(OCSP_BASICRESP),
													&resp_val->responseBytes->response)) {
						PKI_log_err("ERROR while encoding OCSP RESP");
						return ( PKI_ERR );
					}
				}
			} else {
				PKI_log_debug("Test Mode: Signature Not Found!");
			};
		};
	};

	if( conf->debug ) PKI_log_debug ("Response signed ok!");

	/*
	if ( conf->debug ) {
		PKI_MEM *mem = NULL;
		mem = PKI_X509_OCSP_RESP_put_mem ( resp, PKI_DATA_FORMAT_ASN1,
				NULL, NULL, NULL );

		PKI_log_debug("RESP converted -> %d", mem->size );

		URL_put_data ( "/tmp/ocsp-req-2.der", mem, NULL,
			NULL, 0, 0, NULL );

		PKI_X509_OCSP_REQ_put(req, PKI_DATA_FORMAT_ASN1,
				"/tmp/ocsp-req.der", NULL, NULL, NULL );
		PKI_X509_OCSP_RESP_put(resp, PKI_DATA_FORMAT_ASN1,
				"/tmp/ocsp-resp.der", NULL, NULL, NULL );
	}
	*/

end:

	if(thisupd) PKI_TIME_free ( thisupd );
	if(nextupd) PKI_TIME_free ( nextupd );

	return resp;
}

int ocspd_resp_send_socket(int connfd, PKI_X509_OCSP_RESP *r, 
						OCSPD_CONFIG *conf) {

	PKI_TIME *date = NULL;
	PKI_TIME *expire = NULL;
	PKI_MEM  *mem = NULL;

	char buf[1024];

	char http_resp[] =
		"HTTP/1.0 200 OK\r\n"
		"Content-Type: application/ocsp-response\r\n"
 		"Content-Transfer-Encoding: Binary\r\n";

	char *eoh = "\r\n";

	if ( connfd <= 0 ) {
		PKI_log_err("Socket fd is 0!");
		return 0;
	}

	date = PKI_TIME_new( 0 );

	if (( mem = PKI_X509_OCSP_REQ_put_mem ( r, PKI_DATA_FORMAT_ASN1, 
					NULL, NULL, NULL )) == NULL ) {
		PKI_log_err("Memory Error");
		return ( 0 );
	}

	if( conf->set_nextUpdate ) {
		expire = PKI_TIME_new ( (conf->nmin * 60 ) +
				(conf->ndays * 86400 ) );
	} else {
		expire = PKI_TIME_new( 0 );
	}

	date = PKI_TIME_new ( 0 );

	PKI_NET_write ( connfd, http_resp, sizeof(http_resp) - 1);

	sprintf( buf, "Content-Length: %d\r\n", mem->size);
	PKI_NET_write ( connfd, buf, strlen(buf));

	sprintf( buf, "Date: %s\r\n", PKI_TIME_get_parsed (date));
	PKI_NET_write ( connfd, buf, strlen(buf));

	sprintf( buf, "Expires: %s\r\n", PKI_TIME_get_parsed (expire));
	PKI_NET_write ( connfd, buf, strlen(buf));
	PKI_NET_write (connfd, eoh, strlen(eoh));

	PKI_NET_write (connfd, mem->data, mem->size );
	// PKI_NET_write (connfd, "\r\n", 2); <--- Deleted: Caused issues
	//                                         in Firefox/Thunderbird OCSPs
	fflush(NULL);

	if ( date ) PKI_TIME_free ( date );
	if ( expire ) PKI_TIME_free ( expire );

	return PKI_OK;
}

CA_LIST_ENTRY *OCSPD_CA_ENTRY_find ( OCSPD_CONFIG *conf, OCSP_CERTID *cid ) {
	long i, j;
	int ret;

	OCSP_CERTID *b = NULL;
	STACK_OF(CA_ENTRY_CERTID) *a = NULL;
	CA_LIST_ENTRY *ca = NULL;
	CA_ENTRY_CERTID *tmp = NULL;

	b = cid;

	if( (conf == NULL) || ( conf->ca_list == NULL ))
		return NULL;

	for( i = 0; i < PKI_STACK_elements(conf->ca_list); i++ ) {

		// a = conf->ca_list[i]->cid;
		ca = (CA_LIST_ENTRY *) PKI_STACK_get_num( conf->ca_list, i );

		tmp = ca->cid;

		// for( j = 0; j < sk_CA_ENTRY_CERTID_num(a); j++ ) {
		// 	if((tmp = sk_CA_ENTRY_CERTID_value(a,j))==NULL )
		// 		break;

			/* Check for hashes */
			if((ret = ASN1_OCTET_STRING_cmp(tmp->nameHash, 
						b->issuerNameHash)) != 0 ){
				if( conf->debug ) {
					PKI_log_debug("CRL::CA "
					"[%s] nameHash mismatch (%d)", 
					ca->ca_id, ret);
				};
				continue;
				// return NULL;
			}

				if( conf->debug ) {
					PKI_log_debug("CRL::CA "
					"[%s] nameHash OK", ca->ca_id);
				};

			if((ret = ASN1_OCTET_STRING_cmp(tmp->keyHash,
						b->issuerKeyHash)) != 0 ) {
				if( conf->debug ) {
					PKI_log_debug("CRL::CA "
					"[%s] issuerKeyHash mismatch (%d)",
				 	ca->ca_id, ret);
				};

				if( conf->debug ) {
					PKI_log_debug("CRL::CA "
					"[%s] issuerKeyHash OK", ca->ca_id);
				};

				continue;
				// return NULL;
			}
			/* If here we have found it! */
			// if (!ret) return( conf->ca_list[i] );
			if (!ret) return ( ca );
		// }
	}
	/* Here we have not found any suitable CA */
	return(NULL);
}

X509_REVOKED *OCSPD_REVOKED_find (CA_LIST_ENTRY *ca, ASN1_INTEGER *serial) {

	X509_REVOKED *r = NULL;

	int curr = 0;
	int start = 0;
	int cont = 1;
	int end, cmp_val;
	int found = 0;

	/* If no entries are in the list, return directly */
	if( !(ca) || !(ca->crl) || !(ca->crl_list)) return (r);
 
	/* Set the end point to the last one */
	end = sk_X509_REVOKED_num(ca->crl_list) - 1;
	if( end < 0 ) return (r);

	while( cont == 1 ) {
		/* We have not found the entry */
		if( end < start ) break;

		/* Calculate the middle between start and end */
		curr = (int) ((end - start) / 2) + start;

		/* Get the entry from the stack */
		r = sk_X509_REVOKED_value(ca->crl_list, curr);

		/* Compare the two serials */
		cmp_val = ASN1_INTEGER_cmp(r->serialNumber, serial);

		if( cmp_val > 0 ) {
			end = curr - 1;
			continue;
		} else if ( cmp_val < 0 ) {
			start = curr + 1;
			continue;
		} else {
			/* Entry Found ! */
			cont = 0;
			found = 1;
			break;
		}
	}
	if( found )
		return (r);
	else
		return(NULL);

}

